#include "tree.h"

int memcount = 0;

void * d_malloc( size_t size )
   {
      ++memcount;
      return malloc(size);
   }

void d_free(void * memblock )
   {
      --memcount;
      free(memblock);
   }

int getmemcount(void)
   {
      return memcount;
   }


/******************************************************************************
 * Prft ob es in dem Baum tree ein Element mit dem Namen name gibt
 ******************************************************************************/
int isNameIn(t_tree *tree, char *name)
{
 int i;

 if (!tree || !name)
   return 0;

 if (!strcmp(tree->data.name, name) )
   return 1;
 
 for (i=0; i<tree->data.childNodeCount; ++i) 
   {
      if ( isNameIn(tree->next[i], name) )
         return 1;
   }

 return 0;
}


/******************************************************************************
 * Prft, ob ein Node in dem Baum vorhanden ist
 ******************************************************************************/
int isIn(t_tree *tree, t_tree *node)
{
 int i; 

 if (!tree)
   return 0;

 if (tree == node)
   return 1;

 for (i=0; i<tree->data.childNodeCount; ++i) 
   {
      if ( isIn(tree->next[i], node) )
         return 1;
   }

 return 0; 
}


/******************************************************************************
 * Prft, ob ein Node gleich einem anderen Node ist
 ******************************************************************************/
int isEqual(t_tree *node1, t_tree *node2)
{
 if ( node1 == node2                               ||          /* zeiger direkt gleich */
      !strcmp(node1->data.name, node2->data.name)     )        /* name gleich */
   return 1;
 else
   return 0;
}


/******************************************************************************
 * Prft, ob ein Name gltig ist
 ******************************************************************************/
int isName(char *name)
{
 return name && name[0];

}

/******************************************************************************
 * Gibt den bergeordneten Knoten zurck oder Null bei Fehler
 * bei der Wurzel wird die Wurzel selbst zurckgegeben
 ******************************************************************************/
t_tree * getParent(t_tree *tree, t_tree *node)
{
 int i;
 t_tree *t;

 if (!tree || !node)                                           /* falsche Params */
   return NULL;

 if (tree == node)                                             /* wurzel -> kein parent */
   return tree;

 for (i=0; i<tree->data.childNodeCount; ++i)                   /* alle Untertrees durchsuchen */
   {

      if (tree->next[i] == node)                               /* ist die node eine Referenz? */
         return tree;

      if ((t = getParent(tree->next[i], node)) != NULL )       /* durchsuche Unterbaum */
         return t;
   }

 return NULL; 
}



/******************************************************************************
 * Erstellt einen leeren Baum und liefert diesen zurueck
 * Das Funktionsergebnis ist der NULL-Zeiger
 ******************************************************************************/
t_tree* tree_makeEmptyTree(void)
{
 return NULL;
}


/******************************************************************************
 * Erstellt einen Baum und liefert diesen zurueck
 * Die next Referenzen stehen auf NULL (kein Speicher alloziert)
 * Liefert bei Fehler NULL zurueck (wenn name leer oder profit <0)
 ******************************************************************************/
t_tree* tree_makeTree(char* name, float profit)
{
 t_tree * res;

 if (!isName(name) || profit<0.0f)                               /* wenn Parameter ungltig -> Abbruch */
   return tree_makeEmptyTree();

 res = d_malloc(sizeof(t_tree));                                 /* speicher fr element reservieren */
 res->data.name = d_malloc(strlen(name)+1);                      /* speicher fr string reservieren */
 strcpy(res->data.name, name);
 res->data.childNodeCount = 0;
 res->data.profit = profit;
 res->next = NULL;

 return res;
}


/******************************************************************************
 * Zerstoert einen Baum und gibt saemtlichen Speicher des Baumes wieder frei
 * Gibt NULL bei Erfolg  oder bei Misserfolg alten tree zurueck
 * Gibt auch NULL bei Fehler zurck, wenn tree nicht exisitert
 ******************************************************************************/
t_tree* tree_destroyTree(t_tree* tree)
{
 int i;
 
 if (!tree)                                                    /* tree existiert nicht */
   return NULL;
 
 for (i=0; i<tree->data.childNodeCount; ++i)                   /* alle Untertrees zerstren */
   {
      if (tree_destroyTree(tree->next[i]) )
         return tree;                                          /* Fehler -> alten tree zurck */
   }

 d_free(tree->data.name);                                      /* Speicher von Namen freigeben */
 if (tree->next)                                               /* Speicher von vorhandenen Unter-Nodes freigeben */
   d_free(tree->next);
 d_free(tree);                                                 /* das Element freigeben */
 return NULL;
}


/******************************************************************************
 * Fuegt einen einzelnen Knoten vor dem Element 'before' in einen Baum ein.
 * Ein Name darf nur einmal vorkommen !
 * Bei Fehler wird NULL zureuckgeliefert sonst der resultierende Tree
 * (Fehler wenn kein tree, keine before, kein name oder profit<0 oder name 
 * schon vorhanden)
 ******************************************************************************/
t_tree* tree_insertNode(t_tree* tree, t_tree* before, char* name, float profit)
{
 t_tree * neu;
 char   * tname;
 
 if ( !tree                ||                                  /* ungueltige Params -> Abbruch */
      !before              ||
      !isIn(tree,before)   ||
      !isName(name)        ||
      profit<0.0f          ||
      isNameIn(tree, name) )
   return NULL;

 
 neu = tree_makeTree(name, profit);                            /* neues Element erstellen */
 tname = neu->data.name;                                       /* sicher den Zeiger auf den String */
 *neu = *before;                                               /* das neue Element ist das alte Element */

 before->data.childNodeCount = 1;                              /* das alte Element als neues umfunktionieren */
 before->data.name = tname;
 before->data.profit = profit;
 before->next = d_malloc(sizeof(t_tree*)*1);
 before->next[0] = neu;
 return tree;
}

/******************************************************************************
 * Ersetzt eine Node durch eine neue mit den angegebenen Parametern.
 * Die Referenzen der alten Node werden uebernommen.
 * Ein Name darf nur einmal vorkommen !
 * Achtet darauf, den Speicher der ersetzten Node ordentlich freizugeben.
 * Bei Fehler wird NULL zureuckgeliefert. sonst der resultierende Tree
 * (Fehler wenn kein tree, keine node, kein name, profit<0, name schon 
 * vorhanden, node nicht im tree vorhanden)
 ******************************************************************************/
t_tree* tree_replaceNode(t_tree* tree, t_tree* node, char* name, float profit)
{

 if ( !tree                                                    || /* ungueltige Params -> Abbruch */
      !node                                                    ||
      !isName(name)                                            ||
      profit<0.0f                                              ||
      !isIn(tree, node)        )
   return NULL;


 if (!strcmp(node->data.name, name) )                          /* sind Name und neuer Name gleich? */
   {
      node->data.profit = profit;                              /* profit kopieren */
      return tree;
   }
   
 if ( isNameIn(tree, name) )                                   /* Name ist bereits enthalten */
   return NULL;
 
 d_free(node->data.name);                                      /* alten namen freigeben */
 node->data.name = d_malloc(strlen(name)+1);                   /* speicher fr neuen namen reservieren */
 strcpy(node->data.name, name);                                /* namen kopieren */
 node->data.profit = profit;                                   /* profit kopieren */

 return tree;
}

/******************************************************************************
 * Entfernt eine Node aus einem Baum. Alle Referenzen des zu entfernenden 
 * Baumes werden dem darueberliegendem Node angehaengt.
 * Liefert den resultierenden Baum oder NULL bei Fehler zurueck
 * (Fehler bei kein tree, kein node, oder node nicht im Baum)
 * Achtet darauf den Speicher der gelschten Node ordentlich freizugeben.
 ******************************************************************************/
t_tree* tree_removeNode(t_tree* tree, t_tree* node)
{
 t_tree *p, **neu=NULL;
 int i, f=0;

 if ( !tree                ||                                  /* ungueltige Params -> Abbruch */
      !node                ||
      !isIn(tree, node)    ||
      tree==node              )
   return NULL;

 p = getParent(tree, node);

 /* wenn gengend Referenzen da sind */
 if ( (p->data.childNodeCount+node->data.childNodeCount) > 1)
   {  

      /* Speicher fr Referenzen reservieren */
      neu = d_malloc(sizeof(t_tree*) * (p->data.childNodeCount-1+node->data.childNodeCount) );

      /* die alten Referenzen des Parent (auer zu node) wieder reinpacken */
      for (i=0; i<p->data.childNodeCount; ++i)
         {
            if (p->next[i] != node)
               {
                  neu[f] = p->next[i];
                  ++f;
               }
         }

      /* die Referenzen von node dazupacken */
      for (i=0; i<node->data.childNodeCount; ++i)
         {
            neu[f] = node->next[i];
            ++f;
         }
   }

 
 if (p->next)                    /* die alten Refernzen von parent lschen */
   d_free(p->next);
 p->data.childNodeCount = f;
 p->next = neu;                  /* neue Referenzen zuweisen bzw. Null, falls es keine gibt */

 d_free(node->data.name);        /* gebe Text vom zu entfernden Element frei */
 if (node->next)                 /* gebe die Referenzen-Zeiger-Speicher frei */
   d_free(node->next);
 d_free(node);                   /* gebe das Element selbst frei */
 return tree;
}


/******************************************************************************
 * Entfernt alle Elemente des Baumes remTree, die in dem Baum tree vorhanden
 * sind und gibt den neuen remTree zurck
 ******************************************************************************/
t_tree * removeDub(t_tree * tree, t_tree *remTree)
{
 int i;
 t_tree *t;


 if ((t = tree_findName(remTree, tree->data.name)) != NULL )   /* ist das aktuelle El. im remTree? */
   {
      remTree = tree_removeTree(remTree, t);                   /* entferne es aus remTree */
   }
 
 for (i=0; i<tree->data.childNodeCount; ++i)
   {
      remTree = removeDub(tree->next[i], remTree);  
   }


 return remTree;
}

/******************************************************************************
 * Hanegt einer Node in einem Baum einen Tree als zusaetzliche Referenz an.
 * Wenn diese Node vorher keine Referenz hat, hat sie danach eine, wenn sie
 * vorher eine hat, hat sie danach 2,usw.
 * Eine Node darf nur einmal vorkommen !
 * Gibt den resultierenden Baum oder bei Fehler NULL zurueck
 * (Fehler bei kein tree, keine node, kein insertTree, node nicht in tree)
 ******************************************************************************/
t_tree* tree_appendTree(t_tree* tree, t_tree* node, t_tree* insertTree)
{


 if ( !tree                ||                   /* ungueltige Params -> Abbruch */
      !node                ||
      !insertTree          ||
      !isIn(tree, node)       )
   return NULL;

 if (tree==insertTree)
   return tree;


 /* gucken ob im insertTree eine Node ist, die schon im alten baum tree vorhanden ist -> keine doppelten! */
 insertTree = removeDub(tree, insertTree);
 if (!insertTree)       /* kein Element wurde eingefgt */
   return tree;

 node->data.childNodeCount++;                   /* eine Referenz mehr */
 

 if (node->next)
   {
      node->next = realloc(node->next, sizeof(t_tree*) * node->data.childNodeCount);
      node->next[node->data.childNodeCount-1] = insertTree;
   }
 else
   {
      node->next = d_malloc(sizeof(t_tree*) * 1);
      node->next[0] = insertTree;
   }
  
 return tree;
}

/******************************************************************************
 * Ersetzt eine Node durch den angegebenen 'replaceTree'. Die Referenzen 
 * des alten gehen verloren !
 * Achtet darauf, den Speicher der verlorengehenden Referenzen ordentlich
 * zu befreien.
 * Eine Node darf nur einmal vorkommen !
 * Gibt den resultierenden Baum oder bei Fehler NULL zurck
 * (Fehler bei kein tree, kein node, kein replaceTree, wenn node nicht in tree)
 ******************************************************************************/
t_tree* tree_replaceTree(t_tree* tree, t_tree* node, t_tree* replaceTree)
{
 t_tree *p; 

 if ( !tree                ||                   /* ungueltige Params -> Abbruch */
      !node                ||
      !replaceTree         ||
      !isIn(tree, node)       )
   return NULL;

 p = getParent(tree, node);
 tree = tree_removeTree(tree, node);
 tree = tree_appendTree(tree, p, replaceTree);

 return tree;
}

/******************************************************************************
 * Entfernt eine Node mitsamt seinen Referenzen aus einem Baum.
 * Gibt den resultierenden Baum zurueck oder NULL bei Fehler
 * (Fehler wenn kein tree, keine node, node nicht in tree)
 * Achtet darauf, den Speicher der verlorengehenden Referenzen ordentlich
 * zu befreien. 
 ******************************************************************************/
t_tree* tree_removeTree(t_tree* tree, t_tree* node)
{
 t_tree *p, **neu=NULL;
 int i, f=0;

 if ( !tree                ||                                  /* ungueltige Params -> Abbruch */
      !node                ||
      !isIn(tree, node)     )
   return NULL;

 p = getParent(tree, node);

 if (p==node)                                                  /* node ist wurzel -> gesamten Baum killen */
   return tree_destroyTree(tree);

 /* Zeiger auf node-Referenz im parent killen */
 if ( p->data.childNodeCount > 1)                              /* wenn gengend Referenzen da sind */
   {  

      /* Speicher fr Referenzen reservieren */
      neu = d_malloc(sizeof(t_tree) * (p->data.childNodeCount-1) );

      /* die alten Referenzen des Parent (auer zu node) wieder reinpacken */
      for (i=0; i<p->data.childNodeCount; ++i)
         {
            if (p->next[i] != node)
               {
                  neu[f] = p->next[i];
                  ++f;
               }
         }

   }
 if (p->next)                    /* die alten Refernzen von parent lschen */
   d_free(p->next);
 p->data.childNodeCount--;
 p->next = neu;                  /* neue Referenzen zuweisen bzw. Null, falls es keine gibt */
  
 tree_destroyTree(node);         /* baum ab node killen */
 return tree;
}

/******************************************************************************
 * Prueft, ob ein Knoten in einem Baum vorhanden ist
 * Liefert 1 falls vorhanden, sonst 0, bei Fehler -1
 * (Fehler wenn kein tree, keine node)
 ******************************************************************************/
int tree_inTree(t_tree* tree, t_tree* node)
{
 int i;

 if ( !tree || !node )                                 /* ungueltige Params -> Abbruch */
   return -1;


 if (isEqual(tree, node) )
   return 1;

 for (i=0; i<tree->data.childNodeCount; ++i)
   {
      if (tree_inTree(tree->next[i], node) )
         return 1;
   }

 return 0;
}

/******************************************************************************
 * Prueft, ob ein Knoten mit dem Namen 'name' im Baum vorhanden ist und gibt 
 * ihn zurueck oder NULL bei Fehler
 * (Fehler wenn kein tree, kein name)
 ******************************************************************************/
t_tree* tree_findName(t_tree* tree, char* name)
{
 t_tree * res;
 int i;

 if ( !tree || !isName(name) )                                 /* ungueltige Params -> Abbruch */
   return NULL;

 if (!strcmp(tree->data.name, name) )                          /* hat aktueller Node den Namen? */
   return tree;

 for (i=0; i<tree->data.childNodeCount; ++i)                   /* durchlaufe alle Children */
   {
      if ((res = tree_findName(tree->next[i], name)) != NULL ) /* wenn Children den Namen hat */
         return res;                                           /* Zeiger auf Child zurckgeben */
   }

 return NULL;
}


/******************************************************************************
 * addiert den Profit auf das Element node und alle Parent-Nodes
 ******************************************************************************/
t_tree* updateProfit(t_tree *tree, t_tree *node, float profit)
{
 t_tree *p = getParent(tree, node);

 if (p == node)                                                /* Wurzel erreicht? */
   {
      p->data.profit += profit;                                /* Wurzel bekommt gesamten Profit */
      return p;
   }

 node->data.profit += (float) (profit*(1-PROFIT_ANTEIL));      /* aktuelles Element bekommt Profit-Anteil */
 return updateProfit(tree, p, (float) (profit*PROFIT_ANTEIL)); /* Parent Node bekommt auch Profit Anteil */
}


/******************************************************************************
 * Erhoeht den Profit eines Nodes um den uebergebenen profit*(1-PROFIT_ANTEIL).
 * Simuliert also somit ein gemachtes Geschaeft eines Node, der PROFIT_ANTEIL
 * an die Nodes ueber ihm die ihn referenzieren abdruecken muss, d.h. er selbst
 * behaelt nur 1-PROFIT_ANTEIL des angegebenen Profits.
 * Aktualisiert also auch die Profits der Nodes darueber, je unter beruecksich-
 * tigung der PROFIT_ANTEIL Abgabe an Nodes darueber
 * Der oberste muss natuerlich nichts mehr abgeben, behaelt also dann 100%
 * Gibt den resuliterenden Baum zureuck oder NULL bei Fehler
 * (Fehler wenn kein tree, kein name, profit<0, name nicht in tree)
 ******************************************************************************/
t_tree* tree_updateProfit(t_tree* tree, char *name, float profit)
{
 t_tree *t;
 
 if ( !tree                ||                                  /* ungueltige Params -> Abbruch */
      !isName(name)        ||
      profit<0             ||
      !isNameIn(tree, name)   )
   return NULL;

 t = tree_findName(tree, name);                                /* suche Element */
 
 return updateProfit(tree, t, profit);
}

/******************************************************************************
 * Ermittelt den von einem Node erzielten Profit.
 * Gibt den Profit oder -1.0 bei Fehler zureuck
 * (Fehler wenn kein tree, kein name, name nicht in tree)
 ******************************************************************************/
float tree_getNodeProfit(t_tree* tree, char* name)
{
 t_tree *t;
 
 if (tree && name && (t=tree_findName(tree, name)) )           /* gltige Params -> Profit Element bestimmen */
      return t->data.profit;                                   /* Profit zurckgeben */

 return -1.0f;
}


/******************************************************************************
 * Ermittelt den im gesamten Baum erzielten Profit.
 * Gibt den Profit oder -1.0 bei Fehler zureuck
 * (Fehler wenn kein tree)
 ******************************************************************************/
float tree_getTreeProfit(t_tree* tree)
{
 int i;
 float profit;

 if (!tree)                                                    /* ungltiger Baum */
   return -1.0f;

 profit = tree->data.profit;                                   /* Profits des Knotens */

 for (i=0; i<tree->data.childNodeCount; ++i)                   /* durchlaufe alle Unterknoten */
   {
      profit += tree_getTreeProfit(tree->next[i]);             /* Unter Profite addieren */
   }

 return profit;
}

